using System;
using System.Data;
using System.Configuration;
using System.Collections;
using System.Web;
using System.Web.Security;
using System.Web.UI;
using System.Web.UI.WebControls;
using System.Web.UI.WebControls.WebParts;
using System.Web.UI.HtmlControls;
using System.IO;
using System.Web.Caching;
using System.Text.RegularExpressions;

using CMS.GlobalHelper;
using CMS.CMSHelper;
using CMS.FileManager;
using CMS.SiteProvider;
using CMS.URLRewritingEngine;
using CMS.PortalEngine;
using CMS.DataEngine;
using CMS.MediaLibrary;
using CMS.UIControls;
using CMS.DirectoryUtilities;
using CMS.SettingsProvider;

public partial class CMSPages_GetMediaFile : GetFilePage
{
    #region "Advanced settings"

    /// <summary>
    /// Set to false to disable the client caching
    /// </summary>
    protected bool useClientCache = true;

    /// <summary>
    /// Set to 0 if you do not wish to cache large files
    /// </summary>
    protected int largeFilesCacheMinutes = 1;

    #endregion


    #region "Variables"

    protected CMSOutputMediaFile outputFile = null;
    protected MediaFileInfo fileInfo = null;

    #endregion


    #region "Properties"

    /// <summary>
    /// Returns file guid from query.
    /// </summary>
    public Guid FileGuid
    {
        get
        {
            return QueryHelper.GetGuid("fileguid", Guid.Empty);
        }
    }


    /// <summary>
    /// Returns true if file preview is used for sending.
    /// </summary>
    public bool Preview
    {
        get
        {
            return QueryHelper.GetBoolean("preview", false);
        }
    }

    #endregion


    protected void Page_Load(object sender, EventArgs e)
    {
        ConnectionHelper.SetContext("GetMediaFile");

        // Check the site
        if (string.IsNullOrEmpty(CurrentSiteName))
        {
            throw new Exception("[GetMediaFile.aspx]: Site not running.");
        }

        string useCacheItemName = this.CacheItemName;

        #region "Security"

        // Check if user is allowed to see the library file content if required
        bool checkPermissions = SettingsKeyProvider.GetBoolValue(CMSContext.CurrentSiteName + ".CMSCheckMediaFilePermissions");
        if (checkPermissions)
        {
            // If required file was specified
            if (this.FileGuid != Guid.Empty)
            {
                fileInfo = MediaFileInfoProvider.GetMediaFileInfo(this.FileGuid, CMSContext.CurrentSiteName);
                if (fileInfo != null)
                {
                    // Check the library access for the current user and stop file processing if not allowed
                    MediaLibraryInfo mli = MediaLibraryInfoProvider.GetMediaLibraryInfo(fileInfo.FileLibraryID);
                    if (!MediaLibraryInfoProvider.IsUserAuthorizedPerLibrary(mli, "LibraryAccess"))
                    {
                        // Display nothing when necessary
                        SendFile(null);
                        return;
                    }
                }
            }
        }

        #endregion

        // Seek the attachment in cache
        if ((CacheHelper.Contains(useCacheItemName)) && (CacheHelper.CacheImageEnabled(CurrentSiteName)) && (CMSContext.ViewMode == ViewModeEnum.LiveSite))
        {
            // Get OutputFile object and process it.
            outputFile = (CMSOutputMediaFile)CacheHelper.GetItem(useCacheItemName);
        }
        else
        {
            // Get file, process it and store it in cache if caching enabled.
            ProcessFile();

            // Save to cache if enabled and live site mode
            if (CacheHelper.CacheImageEnabled(CurrentSiteName) && (CMSContext.ViewMode == ViewModeEnum.LiveSite))
            {
                int cacheMinutes = CacheHelper.CacheImageMinutes(CurrentSiteName);

                CacheDependency cd = null;
                if (outputFile != null)
                {
                    if (outputFile.MediaFile != null)
                    {
                        // Do not cache too big files
                        if ((outputFile.MediaFile != null) && !CacheHelper.CacheImageAllowed(CurrentSiteName, outputFile.MediaFile.FileSize))
                        {
                            cacheMinutes = largeFilesCacheMinutes;
                        }
                        string[] dependencies = MediaFileInfoProvider.GetDependencyCacheKeys(outputFile.MediaFile, this.Preview);
                        cd = CacheHelper.GetCacheDependency(dependencies);
                    }
                }

                if (cacheMinutes > 0)
                {
                    CacheHelper.Add(useCacheItemName, outputFile, cd, DateTime.Now.AddMinutes(cacheMinutes), Cache.NoSlidingExpiration, CacheHelper.CacheItemPriority, null);
                }
            }
        }

        // Send the data
        SendFile(outputFile);

        ConnectionHelper.ReleaseContext();
    }


    /// <summary>
    /// Processes the file.
    /// </summary>
    protected void ProcessFile()
    {
        if (this.FileGuid != Guid.Empty)
        {
            ProcessFile(this.FileGuid, this.Preview);
        }
    }


    /// <summary>
    /// Processes the specified file.
    /// </summary>
    /// <param name="fileGuid">File guid</param>
    /// <param name="preview">Use preview</param>
    protected void ProcessFile(Guid fileGuid, bool preview)
    {
        // Get the file info if doesn't retrieved yet
        fileInfo = ((fileInfo == null) ? MediaFileInfoProvider.GetMediaFileInfo(fileGuid, CurrentSiteName) : fileInfo);
        if (fileInfo != null)
        {
            if (preview)
            {
                // Get file path
                string path = MediaFileInfoProvider.GetMediaFilePath(fileInfo.FileLibraryID, fileInfo.FilePath);

                // Ensure hidden folder exists
                DirectoryHelper.EnsureDiskPath(Path.GetDirectoryName(path) + "\\" + MediaHelper.GetMediaFileHiddenFolder(CurrentSiteName), Path.GetDirectoryName(path));
                // Get preview file
                string[] files = Directory.GetFiles(Path.GetDirectoryName(path) + "\\" + MediaHelper.GetMediaFileHiddenFolder(CurrentSiteName), MediaHelper.GetPreviewFileName(fileInfo.FileName, fileInfo.FileExtension, ".*", CurrentSiteName));
                if (files.Length > 0)
                {
                    bool resizeImage = (ImageHelper.IsImage(Path.GetExtension(files[0])) && MediaFileInfoProvider.CanResizeImage(files[0], this.Width, this.Height, this.MaxSideSize));

                    // Get the data
                    if ((outputFile == null) || (outputFile.MediaFile == null))
                    {
                        outputFile = new CMSOutputMediaFile(fileInfo, null);
                        outputFile.UsePreview = true;
                        outputFile.Width = this.Width;
                        outputFile.Height = this.Height;
                        outputFile.MaxSideSize = this.MaxSideSize;
                        outputFile.Resized = resizeImage;
                    }
                }
            }
            else
            {
                bool resizeImage = (ImageHelper.IsImage(fileInfo.FileExtension) && MediaFileInfoProvider.CanResizeImage(fileInfo, this.Width, this.Height, this.MaxSideSize));

                // Get the data
                if ((outputFile == null) || (outputFile.MediaFile == null))
                {
                    outputFile = new CMSOutputMediaFile(fileInfo, null);
                    outputFile.Width = this.Width;
                    outputFile.Height = this.Height;
                    outputFile.MaxSideSize = this.MaxSideSize;
                    outputFile.Resized = resizeImage;
                }
            }
        }
    }


    /// <summary>
    /// Sends the given file within response
    /// </summary>
    /// <param name="file">File to send</param>
    protected void SendFile(CMSOutputMediaFile file)
    {
        // Clear response.
        Response.Cookies.Clear();
        Response.Clear();

        Response.Cache.SetRevalidation(HttpCacheRevalidation.AllCaches);

        if (file != null)
        {
            // Prepare etag
            string etag = "";

            // Client caching - only on the live site
            if (useClientCache && (CMSContext.ViewMode == ViewModeEnum.LiveSite) && CacheHelper.CacheImageEnabled(CurrentSiteName))
            {
                // Determine the last modified date and etag sent from the browser
                string currentETag = Request.Headers["If-None-Match"];
                string ifModifiedString = Request.Headers["If-Modified-Since"];
                if ((ifModifiedString != null) && (currentETag == etag))
                {
                    // IE-browsers send something like this:
                    // If-Modified-Since: Wed, 19 Jul 2006 11:19:59 GMT; length=3350
                    DateTime ifModified;
                    if (DateTime.TryParse(ifModifiedString.Split(";".ToCharArray())[0], out ifModified))
                    {
                        // If not changed, let the browser use the cached data
                        if (file.FileModifiedWhen <= ifModified.AddSeconds(1))
                        {
                            Response.StatusCode = (int)System.Net.HttpStatusCode.NotModified;
                            CompleteRequest();
                            return;
                        }
                    }
                }
            }

            // If physical file not present, try to load
            EnsurePhysicalFile(outputFile);

            // If the output data should be cached, return the output data
            bool cacheOutputData = false;
            if (file.MediaFile != null)
            {
                cacheOutputData = CacheHelper.CacheImageAllowed(CurrentSiteName, file.MediaFile.FileSize);
            }

            // Ensure the file data if physical file not present
            if (!file.DataLoaded && (file.PhysicalFile == ""))
            {
                byte[] cachedData = GetCachedOutputData();
                if (file.EnsureData(cachedData))
                {
                    if ((cachedData == null) && cacheOutputData)
                    {
                        SaveOutputDataToCache(file.OutputData, GetOutputDataDependency(file.MediaFile));
                    }
                }
            }

            // Send the file
            if ((file.OutputData != null) || (file.PhysicalFile != ""))
            {
                // Setup the mime type - Fix the special types
                string mimetype = file.MimeType;
                switch (file.MediaFile.FileExtension.ToLower())
                {
                    case ".flv":
                        mimetype = "video/x-flv";
                        break;
                }

                // Prepare response
                Response.ContentType = mimetype;
                SetDisposition(file.MediaFile.FileName + file.MediaFile.FileExtension, file.MediaFile.FileExtension);

                if (useClientCache && (CMSContext.ViewMode == ViewModeEnum.LiveSite) && (CacheHelper.CacheImageEnabled(CurrentSiteName)))
                {
                    DateTime expires = DateTime.Now;

                    // Send last modified header to allow client caching
                    Response.Cache.SetLastModified(file.FileModifiedWhen);

                    Response.Cache.SetCacheability(HttpCacheability.Public);
                    if (DocumentBase.AllowClientCache() && DocumentBase.UseFullClientCache)
                    {
                        expires = DateTime.Now.AddMinutes(CacheHelper.CacheImageMinutes(CurrentSiteName));
                    }

                    Response.Cache.SetExpires(expires);
                    Response.Cache.SetETag(etag);
                }

                // Add the file data
                if ((file.PhysicalFile != "") && (file.OutputData == null))
                {
                    // Stream the file from the file system
                    file.OutputData = WriteFileFix(file.PhysicalFile, cacheOutputData);
                }
                else
                {
                    // Use output data of the file in memory if present
                    WriteBytes(file.OutputData);
                }
            }
            else
            {
                Response.Write("<html></html>");
            }
        }
        else
        {
            Response.Write("<html></html>");
        }

        CompleteRequest();
        //Response.End();
    }


    /// <summary>
    /// Returns the output data dependency based on the given attachment record
    /// </summary>
    /// <param name="mi">MediaFile object</param>
    protected CacheDependency GetOutputDataDependency(MediaFileInfo mi)
    {
        if (mi == null)
        {
            return null;
        }

        return CacheHelper.GetCacheDependency(MediaFileInfoProvider.GetDependencyCacheKeys(mi, this.Preview));
    }


    /// <summary>
    /// Ensures the physical file
    /// </summary>
    /// <param name="file">Output file</param>
    public bool EnsurePhysicalFile(CMSOutputMediaFile file)
    {
        // Try to link to file system
        if ((file != null) && (file.MediaFile != null) && (file.MediaFile.FileID > 0))
        {
            SiteInfo si = SiteInfoProvider.GetSiteInfo(file.MediaFile.FileSiteID);
            if (si != null)
            {
                bool generateThumbnails = ValidationHelper.GetBoolean(SettingsKeyProvider.GetStringValue(si.SiteName + ".CMSGenerateThumbnails"), true);
                string filePath = null;
                if (file.UsePreview)
                {
                    if ((file.Resized) && (generateThumbnails))
                    {
                        filePath = MediaFileInfoProvider.EnsureThumbnailFile(file.MediaFile, file.SiteName, this.Width, this.Height, this.MaxSideSize, true);
                    }
                    else
                    {
                        string hiddenFolder = MediaLibraryInfoProvider.GetMediaLibraryFolderPath(file.MediaFile.FileLibraryID) + "\\" + Path.GetDirectoryName(file.MediaFile.FilePath);
                        string searchPattern = file.MediaFile.FileName + ".*";
                        if (Directory.Exists(hiddenFolder))
                        {
                            string[] preview = Directory.GetFiles(hiddenFolder, searchPattern);
                            if ((preview != null) && (preview.Length > 0))
                            {
                                filePath = Path.GetDirectoryName(file.MediaFile.FilePath) + "\\" + Path.GetFileName(preview[0]);
                            }
                        }

                    }
                }
                else
                {
                    if ((file.Resized) && (generateThumbnails))
                    {
                        filePath = MediaFileInfoProvider.EnsureThumbnailFile(file.MediaFile, file.SiteName, this.Width, this.Height, this.MaxSideSize);
                    }
                    else
                    {
                        filePath = file.MediaFile.FilePath;
                    }
                }

                if (filePath != null)
                {
                    // Link to the physical file
                    file.MediaFile.FilePath = filePath;
                    return true;
                }
            }
        }
        return false;
    }





    /// <summary>
    /// If range in HTTP header is presented returns its value
    /// </summary>
    /// <param name="size">Total size of output data</param>
    /// <returns>Offset in bytes</returns>    
    protected int GetRangeFix(int size)
    {
        if (Request.Headers["Range"] != null)
        {
            // Get range
            string bytesString = RegexHelper.GetRegex("bytes=([0-9]*)-", RegexOptions.None).Match(Request.Headers["Range"]).Result("$1");

            // Range was found
            if (!string.IsNullOrEmpty(bytesString))
            {
                // E-tag checking
                if ((Request.Headers["If-Range"]) != null && (Request.Headers["If-Range"] == this.ETag))
                {
                    // Set code 206 - partial content
                    Response.StatusCode = 206;

                    // Add Content-Range header
                    Response.AddHeader("Content-Range", "bytes " + bytesString + "-" + (size - 1).ToString() + "/" + size.ToString());
                    return int.Parse(bytesString);
                }
            }
        }

        return 0;
    }


    /// <summary>
    /// Streams the data file to the response
    /// </summary>
    /// <param name="filepath">File path</param>
    /// <param name="returnOutputData">If true, output data is returned</param>
    protected byte[] WriteFileFix(string filepath, bool returnOutputData)
    {
        if ((filepath == null) || !File.Exists(filepath))
        {
            return null;
        }

        Stream str = null;

        try
        {
            // Open the file.
            str = new FileStream(filepath, FileMode.Open, FileAccess.Read, FileShare.Read, BUFFER_SIZE);

            // Init the file size
            int size = (int)str.Length;
            int dataToRead = size;

            // Variables for maximum stream checking 
            int checkStreamInterval = 0;
            DateTime downloadBegin = DateTime.Now;

            // How often check maximum stream
            // 0 = never, 1 = always
            if (this.MaxByteStream > 0)
            {
                if (this.MaxByteStream >= BUFFER_SIZE)
                {
                    checkStreamInterval = this.MaxByteStream / BUFFER_SIZE;
                }
                else
                {
                    checkStreamInterval = 1;
                }
            }

            Response.AddHeader("Content-Length", dataToRead.ToString());
            Response.AddHeader("Accept-Ranges", "bytes");

            if (dataToRead > 0)
            {
                // Set request timeout
                Server.ScriptTimeout = AttachmentHelper.ScriptTimeout;

                // Resume downloading checking
                str.Position = GetRangeFix(size);

                // Prepare the buffer
                int bufferSize = BUFFER_SIZE;
                if (returnOutputData || (bufferSize > dataToRead))
                {
                    bufferSize = dataToRead;
                }

                byte[] buffer = new byte[bufferSize];
                int iterationIndex = 0;
                DateTime nextConnectionCheck = DateTime.Now.AddSeconds(CONNECTION_TIMEOUT);

                // Setup begin offset 
                int offset = (int)str.Position;
                dataToRead -= offset;
                int beginPosition = offset;

                // Writing data on output is allowed on begin
                bool writeOnOutput = true;

                // Stream checking index
                int checkStreamIndex = 0;

                try
                {
                    // Stream the content
                    while (dataToRead > 0)
                    {
                        int length = 0;
                        if (writeOnOutput)
                        {
                            // Read the data in buffer.
                            length = str.Read(buffer, 0, bufferSize);
                            if (length <= 0)
                            {
                                // Break if no more data is read
                                break;
                            }
                            offset += length;
                        }

                        // Check maximum stream?
                        if (checkStreamInterval != 0)
                        {
                            // Check stream now?
                            if (checkStreamIndex == checkStreamInterval)
                            {
                                // Count time difference between now and download begin
                                TimeSpan difference = DateTime.Now.Subtract(downloadBegin);
                                // Check if limit was reached
                                if ((offset - beginPosition) > (MaxByteStream * (1 + difference.TotalSeconds)))
                                {
                                    writeOnOutput = false;
                                    System.Threading.Thread.Sleep(STREAM_LIMIT_SLEEP);
                                }
                                else
                                {
                                    writeOnOutput = true;
                                    checkStreamIndex = 0;
                                }
                            }
                            else
                            {
                                checkStreamIndex++;
                            }
                        }

                        // Put data on output 
                        if (writeOnOutput)
                        {
                            Response.OutputStream.Write(buffer, 0, length);
                            Response.Flush();
                        }

                        dataToRead -= length;

                        // Verify that the client is connected (not on the first iteration)
                        if ((iterationIndex++ > CONNECTION_CHECK) && (nextConnectionCheck < DateTime.Now))
                        {
                            iterationIndex = 0;

                            // End the connection if not connected
                            if (Response.IsClientConnected)
                            {
                                nextConnectionCheck = DateTime.Now.AddSeconds(CONNECTION_TIMEOUT);
                            }
                            else
                            {
                                break;
                            }
                        }
                    }
                }
                catch (Exception ex)
                {
                }

                // Return the output data
                if (returnOutputData)
                {
                    return buffer;
                }
            }
        }
        finally
        {
            if (str != null)
            {
                str.Close();
            }
        }

        return null;
    }
}
